# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Any

from ..damlast.daml_lf_1 import DefDataType, Type as DamlType
from .mapper import Context, ValueMapper, convert


class PythonSampleEncoder(ValueMapper):
    def __init__(self, max_depth: int = 6):
        self.max_depth = max_depth

    def _recurse(self, context: "Context", key, item_type):
        if context.depth <= self.max_depth:
            return "..."
        else:
            return convert(self, context.append_path(key), item_type, None)

    def data_record(
        self, context: "Context", dt: "DefDataType", record: "DefDataType.Fields", obj: "Any"
    ) -> "Any":
        return {fld.field: self._recurse(context, fld.field, fld.type) for fld in record.fields}
        pass

    def data_variant(
        self, context: "Context", dt: "DefDataType", variant: "DefDataType.Fields", obj: "Any"
    ) -> "Any":
        return {fld.field: self._recurse(context, fld.field, fld.type) for fld in variant.fields}

    def data_enum(
        self,
        context: "Context",
        dt: "DefDataType",
        enum: "DefDataType.EnumConstructors",
        obj: "Any",
    ) -> "Any":
        return " | ".join(enum.constructors)

    def prim_unit(self, context: "Context", obj: "Any") -> "Any":
        return "{}"

    def prim_bool(self, context: "Context", obj: "Any") -> "Any":
        return "true | false"

    def prim_int64(self, context: "Context", obj: "Any") -> "Any":
        return "int-value"

    def prim_text(self, context: "Context", obj: "Any") -> "Any":
        return "text-value"

    def prim_timestamp(self, context: "Context", obj: "Any") -> "Any":
        return "timestamp-value"

    def prim_party(self, context: "Context", obj: "Any") -> "Any":
        return "party-value"

    def prim_list(self, context: "Context", item_type: "DamlType", obj: "Any") -> "Any":
        return [self._recurse(context, "[]", item_type)]

    def prim_date(self, context: "Context", obj: "Any") -> "Any":
        return "date-value"

    def prim_contract_id(self, context: "Context", item_type: "DamlType", obj: "Any") -> "Any":
        return "contract-id"

    def prim_optional(self, context: "Context", item_type: "DamlType", obj: "Any") -> "Any":
        return ""

    def prim_text_map(self, context: "Context", item_type: "DamlType", obj: "Any") -> "Any":
        return [self._recurse(context, "[]", item_type)]

    def prim_numeric(self, context: "Context", nat: int, obj: "Any") -> "Any":
        pass

    def prim_gen_map(
        self, context: "Context", key_type: "DamlType", value_type: "DamlType", obj: "Any"
    ) -> "Any":
        return {self._recurse(context, "[]", key_type): self._recurse(context, "[]", value_type)}



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Any, NewType

from .basic import to_str

__all__ = ["Party", "to_party"]

Party = NewType("Party", str)


def to_party(o: "Any") -> "Party":
    return Party(to_str(o))



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/value.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/daml/ledger/api/v1/value.proto',
  package='com.daml.ledger.api.v1',
  syntax='proto3',
  serialized_options=b'\n\026com.daml.ledger.api.v1B\017ValueOuterClass\252\002\026Com.Daml.Ledger.Api.V1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\"com/daml/ledger/api/v1/value.proto\x12\x16\x63om.daml.ledger.api.v1\x1a\x1bgoogle/protobuf/empty.proto\"\xa6\x04\n\x05Value\x12\x30\n\x06record\x18\x01 \x01(\x0b\x32\x1e.com.daml.ledger.api.v1.RecordH\x00\x12\x32\n\x07variant\x18\x02 \x01(\x0b\x32\x1f.com.daml.ledger.api.v1.VariantH\x00\x12\x15\n\x0b\x63ontract_id\x18\x03 \x01(\tH\x00\x12,\n\x04list\x18\x04 \x01(\x0b\x32\x1c.com.daml.ledger.api.v1.ListH\x00\x12\x13\n\x05int64\x18\x05 \x01(\x12\x42\x02\x30\x01H\x00\x12\x11\n\x07numeric\x18\x06 \x01(\tH\x00\x12\x0e\n\x04text\x18\x08 \x01(\tH\x00\x12\x17\n\ttimestamp\x18\t \x01(\x10\x42\x02\x30\x01H\x00\x12\x0f\n\x05party\x18\x0b \x01(\tH\x00\x12\x0e\n\x04\x62ool\x18\x0c \x01(\x08H\x00\x12&\n\x04unit\x18\r \x01(\x0b\x32\x16.google.protobuf.EmptyH\x00\x12\x0e\n\x04\x64\x61te\x18\x0e \x01(\x05H\x00\x12\x34\n\x08optional\x18\x0f \x01(\x0b\x32 .com.daml.ledger.api.v1.OptionalH\x00\x12*\n\x03map\x18\x10 \x01(\x0b\x32\x1b.com.daml.ledger.api.v1.MapH\x00\x12,\n\x04\x65num\x18\x11 \x01(\x0b\x32\x1c.com.daml.ledger.api.v1.EnumH\x00\x12\x31\n\x07gen_map\x18\x12 \x01(\x0b\x32\x1e.com.daml.ledger.api.v1.GenMapH\x00\x42\x05\n\x03Sum\"t\n\x06Record\x12\x35\n\trecord_id\x18\x01 \x01(\x0b\x32\".com.daml.ledger.api.v1.Identifier\x12\x33\n\x06\x66ields\x18\x02 \x03(\x0b\x32#.com.daml.ledger.api.v1.RecordField\"J\n\x0bRecordField\x12\r\n\x05label\x18\x01 \x01(\t\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.Value\"P\n\nIdentifier\x12\x12\n\npackage_id\x18\x01 \x01(\t\x12\x13\n\x0bmodule_name\x18\x03 \x01(\t\x12\x13\n\x0b\x65ntity_name\x18\x04 \x01(\tJ\x04\x08\x02\x10\x03\"\x84\x01\n\x07Variant\x12\x36\n\nvariant_id\x18\x01 \x01(\x0b\x32\".com.daml.ledger.api.v1.Identifier\x12\x13\n\x0b\x63onstructor\x18\x02 \x01(\t\x12,\n\x05value\x18\x03 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.Value\"P\n\x04\x45num\x12\x33\n\x07\x65num_id\x18\x01 \x01(\x0b\x32\".com.daml.ledger.api.v1.Identifier\x12\x13\n\x0b\x63onstructor\x18\x02 \x01(\t\"7\n\x04List\x12/\n\x08\x65lements\x18\x01 \x03(\x0b\x32\x1d.com.daml.ledger.api.v1.Value\"8\n\x08Optional\x12,\n\x05value\x18\x01 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.Value\"}\n\x03Map\x12\x32\n\x07\x65ntries\x18\x01 \x03(\x0b\x32!.com.daml.ledger.api.v1.Map.Entry\x1a\x42\n\x05\x45ntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.Value\"\xa2\x01\n\x06GenMap\x12\x35\n\x07\x65ntries\x18\x01 \x03(\x0b\x32$.com.daml.ledger.api.v1.GenMap.Entry\x1a\x61\n\x05\x45ntry\x12*\n\x03key\x18\x01 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.Value\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueBB\n\x16\x63om.daml.ledger.api.v1B\x0fValueOuterClass\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,])




_VALUE = _descriptor.Descriptor(
  name='Value',
  full_name='com.daml.ledger.api.v1.Value',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='record', full_name='com.daml.ledger.api.v1.Value.record', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='variant', full_name='com.daml.ledger.api.v1.Value.variant', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='contract_id', full_name='com.daml.ledger.api.v1.Value.contract_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='list', full_name='com.daml.ledger.api.v1.Value.list', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='int64', full_name='com.daml.ledger.api.v1.Value.int64', index=4,
      number=5, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'0\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='numeric', full_name='com.daml.ledger.api.v1.Value.numeric', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text', full_name='com.daml.ledger.api.v1.Value.text', index=6,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='com.daml.ledger.api.v1.Value.timestamp', index=7,
      number=9, type=16, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'0\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='party', full_name='com.daml.ledger.api.v1.Value.party', index=8,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bool', full_name='com.daml.ledger.api.v1.Value.bool', index=9,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='unit', full_name='com.daml.ledger.api.v1.Value.unit', index=10,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='date', full_name='com.daml.ledger.api.v1.Value.date', index=11,
      number=14, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='optional', full_name='com.daml.ledger.api.v1.Value.optional', index=12,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='map', full_name='com.daml.ledger.api.v1.Value.map', index=13,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='enum', full_name='com.daml.ledger.api.v1.Value.enum', index=14,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='gen_map', full_name='com.daml.ledger.api.v1.Value.gen_map', index=15,
      number=18, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Sum', full_name='com.daml.ledger.api.v1.Value.Sum',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=92,
  serialized_end=642,
)


_RECORD = _descriptor.Descriptor(
  name='Record',
  full_name='com.daml.ledger.api.v1.Record',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='record_id', full_name='com.daml.ledger.api.v1.Record.record_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fields', full_name='com.daml.ledger.api.v1.Record.fields', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=644,
  serialized_end=760,
)


_RECORDFIELD = _descriptor.Descriptor(
  name='RecordField',
  full_name='com.daml.ledger.api.v1.RecordField',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='label', full_name='com.daml.ledger.api.v1.RecordField.label', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='com.daml.ledger.api.v1.RecordField.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=762,
  serialized_end=836,
)


_IDENTIFIER = _descriptor.Descriptor(
  name='Identifier',
  full_name='com.daml.ledger.api.v1.Identifier',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='package_id', full_name='com.daml.ledger.api.v1.Identifier.package_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='module_name', full_name='com.daml.ledger.api.v1.Identifier.module_name', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='entity_name', full_name='com.daml.ledger.api.v1.Identifier.entity_name', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=838,
  serialized_end=918,
)


_VARIANT = _descriptor.Descriptor(
  name='Variant',
  full_name='com.daml.ledger.api.v1.Variant',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='variant_id', full_name='com.daml.ledger.api.v1.Variant.variant_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='constructor', full_name='com.daml.ledger.api.v1.Variant.constructor', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='com.daml.ledger.api.v1.Variant.value', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=921,
  serialized_end=1053,
)


_ENUM = _descriptor.Descriptor(
  name='Enum',
  full_name='com.daml.ledger.api.v1.Enum',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='enum_id', full_name='com.daml.ledger.api.v1.Enum.enum_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='constructor', full_name='com.daml.ledger.api.v1.Enum.constructor', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1055,
  serialized_end=1135,
)


_LIST = _descriptor.Descriptor(
  name='List',
  full_name='com.daml.ledger.api.v1.List',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='elements', full_name='com.daml.ledger.api.v1.List.elements', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1137,
  serialized_end=1192,
)


_OPTIONAL = _descriptor.Descriptor(
  name='Optional',
  full_name='com.daml.ledger.api.v1.Optional',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='com.daml.ledger.api.v1.Optional.value', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1194,
  serialized_end=1250,
)


_MAP_ENTRY = _descriptor.Descriptor(
  name='Entry',
  full_name='com.daml.ledger.api.v1.Map.Entry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='com.daml.ledger.api.v1.Map.Entry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='com.daml.ledger.api.v1.Map.Entry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1311,
  serialized_end=1377,
)

_MAP = _descriptor.Descriptor(
  name='Map',
  full_name='com.daml.ledger.api.v1.Map',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='entries', full_name='com.daml.ledger.api.v1.Map.entries', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_MAP_ENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1252,
  serialized_end=1377,
)


_GENMAP_ENTRY = _descriptor.Descriptor(
  name='Entry',
  full_name='com.daml.ledger.api.v1.GenMap.Entry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='com.daml.ledger.api.v1.GenMap.Entry.key', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='com.daml.ledger.api.v1.GenMap.Entry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1445,
  serialized_end=1542,
)

_GENMAP = _descriptor.Descriptor(
  name='GenMap',
  full_name='com.daml.ledger.api.v1.GenMap',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='entries', full_name='com.daml.ledger.api.v1.GenMap.entries', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_GENMAP_ENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1380,
  serialized_end=1542,
)

_VALUE.fields_by_name['record'].message_type = _RECORD
_VALUE.fields_by_name['variant'].message_type = _VARIANT
_VALUE.fields_by_name['list'].message_type = _LIST
_VALUE.fields_by_name['unit'].message_type = google_dot_protobuf_dot_empty__pb2._EMPTY
_VALUE.fields_by_name['optional'].message_type = _OPTIONAL
_VALUE.fields_by_name['map'].message_type = _MAP
_VALUE.fields_by_name['enum'].message_type = _ENUM
_VALUE.fields_by_name['gen_map'].message_type = _GENMAP
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['record'])
_VALUE.fields_by_name['record'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['variant'])
_VALUE.fields_by_name['variant'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['contract_id'])
_VALUE.fields_by_name['contract_id'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['list'])
_VALUE.fields_by_name['list'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['int64'])
_VALUE.fields_by_name['int64'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['numeric'])
_VALUE.fields_by_name['numeric'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['text'])
_VALUE.fields_by_name['text'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['timestamp'])
_VALUE.fields_by_name['timestamp'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['party'])
_VALUE.fields_by_name['party'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['bool'])
_VALUE.fields_by_name['bool'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['unit'])
_VALUE.fields_by_name['unit'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['date'])
_VALUE.fields_by_name['date'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['optional'])
_VALUE.fields_by_name['optional'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['map'])
_VALUE.fields_by_name['map'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['enum'])
_VALUE.fields_by_name['enum'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_VALUE.oneofs_by_name['Sum'].fields.append(
  _VALUE.fields_by_name['gen_map'])
_VALUE.fields_by_name['gen_map'].containing_oneof = _VALUE.oneofs_by_name['Sum']
_RECORD.fields_by_name['record_id'].message_type = _IDENTIFIER
_RECORD.fields_by_name['fields'].message_type = _RECORDFIELD
_RECORDFIELD.fields_by_name['value'].message_type = _VALUE
_VARIANT.fields_by_name['variant_id'].message_type = _IDENTIFIER
_VARIANT.fields_by_name['value'].message_type = _VALUE
_ENUM.fields_by_name['enum_id'].message_type = _IDENTIFIER
_LIST.fields_by_name['elements'].message_type = _VALUE
_OPTIONAL.fields_by_name['value'].message_type = _VALUE
_MAP_ENTRY.fields_by_name['value'].message_type = _VALUE
_MAP_ENTRY.containing_type = _MAP
_MAP.fields_by_name['entries'].message_type = _MAP_ENTRY
_GENMAP_ENTRY.fields_by_name['key'].message_type = _VALUE
_GENMAP_ENTRY.fields_by_name['value'].message_type = _VALUE
_GENMAP_ENTRY.containing_type = _GENMAP
_GENMAP.fields_by_name['entries'].message_type = _GENMAP_ENTRY
DESCRIPTOR.message_types_by_name['Value'] = _VALUE
DESCRIPTOR.message_types_by_name['Record'] = _RECORD
DESCRIPTOR.message_types_by_name['RecordField'] = _RECORDFIELD
DESCRIPTOR.message_types_by_name['Identifier'] = _IDENTIFIER
DESCRIPTOR.message_types_by_name['Variant'] = _VARIANT
DESCRIPTOR.message_types_by_name['Enum'] = _ENUM
DESCRIPTOR.message_types_by_name['List'] = _LIST
DESCRIPTOR.message_types_by_name['Optional'] = _OPTIONAL
DESCRIPTOR.message_types_by_name['Map'] = _MAP
DESCRIPTOR.message_types_by_name['GenMap'] = _GENMAP
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Value = _reflection.GeneratedProtocolMessageType('Value', (_message.Message,), {
  'DESCRIPTOR' : _VALUE,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Value)
  })
_sym_db.RegisterMessage(Value)

Record = _reflection.GeneratedProtocolMessageType('Record', (_message.Message,), {
  'DESCRIPTOR' : _RECORD,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Record)
  })
_sym_db.RegisterMessage(Record)

RecordField = _reflection.GeneratedProtocolMessageType('RecordField', (_message.Message,), {
  'DESCRIPTOR' : _RECORDFIELD,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.RecordField)
  })
_sym_db.RegisterMessage(RecordField)

Identifier = _reflection.GeneratedProtocolMessageType('Identifier', (_message.Message,), {
  'DESCRIPTOR' : _IDENTIFIER,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Identifier)
  })
_sym_db.RegisterMessage(Identifier)

Variant = _reflection.GeneratedProtocolMessageType('Variant', (_message.Message,), {
  'DESCRIPTOR' : _VARIANT,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Variant)
  })
_sym_db.RegisterMessage(Variant)

Enum = _reflection.GeneratedProtocolMessageType('Enum', (_message.Message,), {
  'DESCRIPTOR' : _ENUM,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Enum)
  })
_sym_db.RegisterMessage(Enum)

List = _reflection.GeneratedProtocolMessageType('List', (_message.Message,), {
  'DESCRIPTOR' : _LIST,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.List)
  })
_sym_db.RegisterMessage(List)

Optional = _reflection.GeneratedProtocolMessageType('Optional', (_message.Message,), {
  'DESCRIPTOR' : _OPTIONAL,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Optional)
  })
_sym_db.RegisterMessage(Optional)

Map = _reflection.GeneratedProtocolMessageType('Map', (_message.Message,), {

  'Entry' : _reflection.GeneratedProtocolMessageType('Entry', (_message.Message,), {
    'DESCRIPTOR' : _MAP_ENTRY,
    '__module__' : 'com.daml.ledger.api.v1.value_pb2'
    # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Map.Entry)
    })
  ,
  'DESCRIPTOR' : _MAP,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Map)
  })
_sym_db.RegisterMessage(Map)
_sym_db.RegisterMessage(Map.Entry)

GenMap = _reflection.GeneratedProtocolMessageType('GenMap', (_message.Message,), {

  'Entry' : _reflection.GeneratedProtocolMessageType('Entry', (_message.Message,), {
    'DESCRIPTOR' : _GENMAP_ENTRY,
    '__module__' : 'com.daml.ledger.api.v1.value_pb2'
    # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.GenMap.Entry)
    })
  ,
  'DESCRIPTOR' : _GENMAP,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.GenMap)
  })
_sym_db.RegisterMessage(GenMap)
_sym_db.RegisterMessage(GenMap.Entry)


DESCRIPTOR._options = None
_VALUE.fields_by_name['int64']._options = None
_VALUE.fields_by_name['timestamp']._options = None
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import json
from typing import Any

from ..damlast.daml_lf_1 import DefDataType, Type
from ..damlast.util import find_variant_type
from ..prim import (
    date_to_str,
    datetime_to_str,
    decimal_to_str,
    to_bool,
    to_decimal,
    to_int,
    to_party,
    to_str,
    to_variant,
)
from .context import Context
from .mapper import ValueMapper

__all__ = ["FlatStringMapper", "ArrayStringMapper"]


def maybe_parenthesize(value: str) -> str:
    return "(" + value + ")" if " " in value else value


class StringMapperBase(ValueMapper):
    """
    Return string representations of all types in a format that looks closest to their
    representation in DAML.
    """

    def data_record(
        self, context: "Context", dt: "DefDataType", record: "DefDataType.Fields", obj: "Any"
    ) -> "Any":
        raise NotImplementedError("StringMapperBase.data_record has no implementation")

    def data_variant(
        self, context: "Context", dt: "DefDataType", variant: "DefDataType.Fields", obj: "Any"
    ) -> "Any":
        ctor, value = to_variant(obj)
        obj_type = find_variant_type(dt, variant, ctor)

        value_str = context.append_path(ctor, mapper=FLAT).convert(obj_type, value)
        return f"{ctor} ({maybe_parenthesize(value_str)})"

    def data_enum(
        self,
        context: "Context",
        dt: "DefDataType",
        enum: "DefDataType.EnumConstructors",
        obj: "Any",
    ) -> "Any":
        return context.value_validate_enum(obj, enum)

    def prim_unit(self, context: "Context", obj: "Any") -> "Any":
        return "()"

    def prim_bool(self, context: "Context", obj: "Any") -> "Any":
        return "true" if to_bool(obj) else "false"

    def prim_int64(self, context: "Context", obj: "Any") -> "Any":
        return str(to_int(obj))

    def prim_text(self, context: "Context", obj: "Any") -> "Any":
        """
        Return text formatted as a string, with double quotes.
        """
        return json.dumps(to_str(obj))

    def prim_timestamp(self, context: "Context", obj: "Any") -> "Any":
        return datetime_to_str(obj)

    def prim_party(self, context: "Context", obj: "Any") -> "Any":
        """
        Return a Party literal formatted as a string, with single quotes.
        """
        return repr(to_party(obj))

    def prim_list(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        return "[" + ", ".join(context.convert_list(item_type, obj, mapper=FLAT)) + "]"

    def prim_date(self, context: "Context", obj: "Any") -> "Any":
        return date_to_str(obj)

    def prim_contract_id(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        return context.convert_contract_id(item_type, obj).value

    def prim_optional(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        if obj is not None:
            value_str = context.convert_optional(item_type, obj, mapper=FLAT)
            if " " in value_str:
                return f"Some ({value_str})"
            else:
                return f"Some {value_str}"
        else:
            return "None"

    def prim_text_map(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        return (
            "{"
            + ", ".join(
                json.dumps(key) + ": " + value
                for key, value in context.convert_text_map(item_type, obj, mapper=FLAT).items()
            )
            + "}"
        )

    def prim_numeric(self, context: "Context", nat: int, obj: "Any") -> "Any":
        return decimal_to_str(to_decimal(obj))

    def prim_gen_map(
        self, context: "Context", key_type: "Type", value_type: "Type", obj: "Any"
    ) -> "Any":
        return (
            "{"
            + ", ".join(
                key + ": " + value
                for key, value in context.convert_gen_map(
                    key_type, value_type, obj, mapper=FLAT
                ).items()
            )
            + "}"
        )


class ArrayStringMapper(StringMapperBase):
    def data_record(
        self, context: "Context", dt: "DefDataType", record: "DefDataType.Fields", obj: "Any"
    ) -> "Any":
        """
        Return the fields of this record as an array of strings.
        """
        elems = []
        for field in record.fields:
            val = context.convert(field.type, obj[field.field])
            if isinstance(val, str):
                elems.append(val)
            else:
                elems.extend(val)
        return elems


class FlatStringMapper(StringMapperBase):
    def data_record(
        self, context: "Context", dt: "DefDataType", record: "DefDataType.Fields", obj: "Any"
    ) -> "Any":
        if record.fields:
            return "... with " + "; ".join(
                f"{field.field}: {context.convert(field.type, obj[field.field])}"
                for field in record.fields
            )
        else:
            return "..."


FLAT = FlatStringMapper()



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/admin/package_management_service.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/daml/ledger/api/v1/admin/package_management_service.proto',
  package='com.daml.ledger.api.v1.admin',
  syntax='proto3',
  serialized_options=b'\n\034com.daml.ledger.api.v1.adminB\"PackageManagementServiceOuterClass\252\002\034Com.Daml.Ledger.Api.V1.Admin',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n=com/daml/ledger/api/v1/admin/package_management_service.proto\x12\x1c\x63om.daml.ledger.api.v1.admin\x1a\x1fgoogle/protobuf/timestamp.proto\"\x1a\n\x18ListKnownPackagesRequest\"b\n\x19ListKnownPackagesResponse\x12\x45\n\x0fpackage_details\x18\x01 \x03(\x0b\x32,.com.daml.ledger.api.v1.admin.PackageDetails\"\x87\x01\n\x0ePackageDetails\x12\x12\n\npackage_id\x18\x01 \x01(\t\x12\x14\n\x0cpackage_size\x18\x02 \x01(\x04\x12/\n\x0bknown_since\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x1a\n\x12source_description\x18\x04 \x01(\t\"?\n\x14UploadDarFileRequest\x12\x10\n\x08\x64\x61r_file\x18\x01 \x01(\x0c\x12\x15\n\rsubmission_id\x18\x02 \x01(\t\"\x17\n\x15UploadDarFileResponse2\x9b\x02\n\x18PackageManagementService\x12\x84\x01\n\x11ListKnownPackages\x12\x36.com.daml.ledger.api.v1.admin.ListKnownPackagesRequest\x1a\x37.com.daml.ledger.api.v1.admin.ListKnownPackagesResponse\x12x\n\rUploadDarFile\x12\x32.com.daml.ledger.api.v1.admin.UploadDarFileRequest\x1a\x33.com.daml.ledger.api.v1.admin.UploadDarFileResponseBa\n\x1c\x63om.daml.ledger.api.v1.adminB\"PackageManagementServiceOuterClass\xaa\x02\x1c\x43om.Daml.Ledger.Api.V1.Adminb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])




_LISTKNOWNPACKAGESREQUEST = _descriptor.Descriptor(
  name='ListKnownPackagesRequest',
  full_name='com.daml.ledger.api.v1.admin.ListKnownPackagesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=128,
  serialized_end=154,
)


_LISTKNOWNPACKAGESRESPONSE = _descriptor.Descriptor(
  name='ListKnownPackagesResponse',
  full_name='com.daml.ledger.api.v1.admin.ListKnownPackagesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='package_details', full_name='com.daml.ledger.api.v1.admin.ListKnownPackagesResponse.package_details', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=156,
  serialized_end=254,
)


_PACKAGEDETAILS = _descriptor.Descriptor(
  name='PackageDetails',
  full_name='com.daml.ledger.api.v1.admin.PackageDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='package_id', full_name='com.daml.ledger.api.v1.admin.PackageDetails.package_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='package_size', full_name='com.daml.ledger.api.v1.admin.PackageDetails.package_size', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='known_since', full_name='com.daml.ledger.api.v1.admin.PackageDetails.known_since', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source_description', full_name='com.daml.ledger.api.v1.admin.PackageDetails.source_description', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=257,
  serialized_end=392,
)


_UPLOADDARFILEREQUEST = _descriptor.Descriptor(
  name='UploadDarFileRequest',
  full_name='com.daml.ledger.api.v1.admin.UploadDarFileRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='dar_file', full_name='com.daml.ledger.api.v1.admin.UploadDarFileRequest.dar_file', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='submission_id', full_name='com.daml.ledger.api.v1.admin.UploadDarFileRequest.submission_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=394,
  serialized_end=457,
)


_UPLOADDARFILERESPONSE = _descriptor.Descriptor(
  name='UploadDarFileResponse',
  full_name='com.daml.ledger.api.v1.admin.UploadDarFileResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=459,
  serialized_end=482,
)

_LISTKNOWNPACKAGESRESPONSE.fields_by_name['package_details'].message_type = _PACKAGEDETAILS
_PACKAGEDETAILS.fields_by_name['known_since'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
DESCRIPTOR.message_types_by_name['ListKnownPackagesRequest'] = _LISTKNOWNPACKAGESREQUEST
DESCRIPTOR.message_types_by_name['ListKnownPackagesResponse'] = _LISTKNOWNPACKAGESRESPONSE
DESCRIPTOR.message_types_by_name['PackageDetails'] = _PACKAGEDETAILS
DESCRIPTOR.message_types_by_name['UploadDarFileRequest'] = _UPLOADDARFILEREQUEST
DESCRIPTOR.message_types_by_name['UploadDarFileResponse'] = _UPLOADDARFILERESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ListKnownPackagesRequest = _reflection.GeneratedProtocolMessageType('ListKnownPackagesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTKNOWNPACKAGESREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.admin.package_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.ListKnownPackagesRequest)
  })
_sym_db.RegisterMessage(ListKnownPackagesRequest)

ListKnownPackagesResponse = _reflection.GeneratedProtocolMessageType('ListKnownPackagesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTKNOWNPACKAGESRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.admin.package_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.ListKnownPackagesResponse)
  })
_sym_db.RegisterMessage(ListKnownPackagesResponse)

PackageDetails = _reflection.GeneratedProtocolMessageType('PackageDetails', (_message.Message,), {
  'DESCRIPTOR' : _PACKAGEDETAILS,
  '__module__' : 'com.daml.ledger.api.v1.admin.package_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.PackageDetails)
  })
_sym_db.RegisterMessage(PackageDetails)

UploadDarFileRequest = _reflection.GeneratedProtocolMessageType('UploadDarFileRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADDARFILEREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.admin.package_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.UploadDarFileRequest)
  })
_sym_db.RegisterMessage(UploadDarFileRequest)

UploadDarFileResponse = _reflection.GeneratedProtocolMessageType('UploadDarFileResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADDARFILERESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.admin.package_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.UploadDarFileResponse)
  })
_sym_db.RegisterMessage(UploadDarFileResponse)


DESCRIPTOR._options = None

_PACKAGEMANAGEMENTSERVICE = _descriptor.ServiceDescriptor(
  name='PackageManagementService',
  full_name='com.daml.ledger.api.v1.admin.PackageManagementService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=485,
  serialized_end=768,
  methods=[
  _descriptor.MethodDescriptor(
    name='ListKnownPackages',
    full_name='com.daml.ledger.api.v1.admin.PackageManagementService.ListKnownPackages',
    index=0,
    containing_service=None,
    input_type=_LISTKNOWNPACKAGESREQUEST,
    output_type=_LISTKNOWNPACKAGESRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UploadDarFile',
    full_name='com.daml.ledger.api.v1.admin.PackageManagementService.UploadDarFile',
    index=1,
    containing_service=None,
    input_type=_UPLOADDARFILEREQUEST,
    output_type=_UPLOADDARFILERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_PACKAGEMANAGEMENTSERVICE)

DESCRIPTOR.services_by_name['PackageManagementService'] = _PACKAGEMANAGEMENTSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Optional

from ..damlast.daml_lf_1 import (
    BuiltinFunction,
    Expr,
    FieldWithType,
    PrimCon,
    PrimLit,
    PrimType,
    Type,
    ValName,
)
from ..damlast.visitor import ExprVisitor, IdentityTypeVisitor
from ..model.types_store import PackageStore
from ..util.typing import safe_cast


class TypeComputer(ExprVisitor[Type], IdentityTypeVisitor):
    """
    Evaluate the :class:`Type` of an :class:`Expr`.
    """

    def __init__(self, store: "PackageStore"):
        self.store = safe_cast(PackageStore, store)

    def resolve_type(self, var: str) -> "Optional[Type]":
        return None

    def visit_expr_var(self, var: str) -> "Type":
        pass

    def visit_expr_val(self, val: "ValName") -> "Type":
        vref = self.store.resolve_value_reference(val)
        if vref is not None:
            return self.visit_expr(vref)
        else:
            raise RuntimeError(f"unknown val name: {val}")

    def visit_expr_builtin(self, builtin: "BuiltinFunction") -> "Type":
        pass

    def visit_expr_prim_con(self, prim_con: "PrimCon") -> "Type":
        if PrimCon.CON_TRUE == prim_con or PrimCon.CON_FALSE == prim_con:
            return Type(prim=Type.Prim(PrimType.BOOL, args=()))
        else:
            return Type(prim=Type.Prim(PrimType.UNIT, args=()))

    def visit_expr_prim_lit(self, prim_lit: "PrimLit") -> "Type":
        if prim_lit.int64 is not None:
            return Type(prim=Type.Prim(PrimType.INT64, args=()))
        elif prim_lit.date is not None:
            return Type(prim=Type.Prim(PrimType.DATE, args=()))
        elif prim_lit.timestamp is not None:
            return Type(prim=Type.Prim(PrimType.TIMESTAMP, args=()))
        elif prim_lit.decimal is not None:
            return Type(prim=Type.Prim(PrimType.DECIMAL, args=()))
        elif prim_lit.party is not None:
            return Type(prim=Type.Prim(PrimType.PARTY, args=()))
        elif prim_lit.text is not None:
            return Type(prim=Type.Prim(PrimType.TEXT, args=()))
        else:
            raise RuntimeError("unknown prim lit type")

    def visit_expr_rec_con(self, rec_con: "Expr.RecCon") -> "Type":
        return self.visit_type(Type(con=rec_con.tycon))

    def visit_expr_rec_proj(self, rec_proj: "Expr.RecProj") -> "Type":
        rec_type = self.visit_type_con(rec_proj.tycon)
        old_type = self.store.resolve_type_reference(rec_type.con.tycon)
        for n, v in old_type.as_args_list():
            if n == rec_proj.field:
                return get_new_type(v)

    def visit_expr_variant_con(self, variant_con: "Expr.VariantCon") -> "Type":
        return self.visit_type(Type(con=variant_con.tycon))

    def visit_expr_struct_con(self, struct_con: "Expr.StructCon") -> "Type":
        return Type(
            tuple=Type.Tuple(
                [FieldWithType(fwe.field, self.visit_expr(fwe.expr)) for fwe in struct_con.fields]
            )
        )

    def visit_expr_struct_proj(self, struct_proj: "Expr.StructProj") -> "Type":
        pass

    def visit_expr_app(self, app: "Expr.App") -> "Type":
        fun_type = self.visit_expr(app.fun)
        if fun_type.prim.prim == PrimType.ARROW:
            t = fun_type
            arg_count = len(app.args)
            while arg_count > 0:
                arg_count -= 1
                t = t.prim.args[1]
            return t
        else:
            raise ValueError(f"unexpected type of app body: {fun_type}")

    def visit_expr_ty_app(self, ty_app: "Expr.TyApp") -> "Type":
        core_type = self.visit_expr(ty_app.expr)
        if ty_app.types:

            def unexpected_type_error(*_, **__):
                raise RuntimeError("Unexpected tyapp")

            return core_type.Sum_match(
                lambda var: Type(var=Type.Var(var=var.var, args=var.args + ty_app.types)),
                lambda con: Type(con=Type.Con(tycon=con.tycon, args=con.args + ty_app.types)),
                lambda prim: Type(prim=Type.Prim(prim=prim.prim, args=prim.args + ty_app.types)),
                lambda forall: unexpected_type_error,
                lambda tuple: unexpected_type_error,
            )
        return core_type

    def visit_expr_abs(self, abs_: "Expr.Abs") -> "Type":
        t = self.visit_expr(abs_.body)
        for param in reversed(abs_.param):
            t = arrow_type(param.type, t)
        return t

    def visit_expr_ty_abs(self, ty_abs: "Expr.TyAbs") -> "Type":
        return "..."
        # scope = self.with_type_var() ty_abs.param[0].var
        # return scope.visit_expr(ty_abs.body)

    def visit_expr_case(self, case: "Case") -> "Type":
        # TODO: Probably not right
        return self.visit_expr(case.alts[0].body)

    def visit_expr_let(self, let: "Block") -> "Type":
        return self.visit_expr(let.body)

    def visit_expr_nil(self, nil: "Expr.Nil") -> "Type":
        return Type(prim=Type.Prim(PrimType.LIST, [self.visit_type(nil.type)]))

    def visit_expr_cons(self, cons: "Expr.Cons") -> "Type":
        return Type(prim=Type.Prim(PrimType.LIST, [self.visit_type(cons.type)]))

    def visit_expr_update(self, update: "Update") -> "Type":
        raise Exception("Update not yet supported")

    def visit_expr_scenario(self, scenario: "Scenario") -> "Type":
        raise Exception("Scenario not yet supported")

    def visit_expr_rec_upd(self, rec_upd: "Expr.RecUpd") -> "Type":
        raise Exception("Update not yet supported")

    def visit_expr_struct_upd(self, struct_upd: "Expr.StructUpd") -> "Type":
        raise Exception("Update not yet supported")

    def visit_expr_optional_none(self, optional_none: "Expr.OptionalNone") -> "Type":
        return Type(prim=Type.Prim(PrimType.OPTIONAL, [self.visit_type(optional_none.type)]))

    def visit_expr_optional_some(self, optional_some: "Expr.OptionalSome") -> "Type":
        return Type(prim=Type.Prim(PrimType.OPTIONAL, [self.visit_type(optional_some.type)]))

    def visit_type_var(self, var: "Type.Var") -> "Type":
        resolved_type = self.resolve_type(var.var)
        if resolved_type is not None:
            return self.visit_type(resolved_type)
        else:
            return Type(var=var)


def arrow_type(input: "Type", output: "Type") -> "Type":
    return Type(prim=Type.Prim(PrimType.ARROW, (input, output)))



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/transaction_filter.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import value_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/daml/ledger/api/v1/transaction_filter.proto',
  package='com.daml.ledger.api.v1',
  syntax='proto3',
  serialized_options=b'\n\026com.daml.ledger.api.v1B\033TransactionFilterOuterClass\252\002\026Com.Daml.Ledger.Api.V1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n/com/daml/ledger/api/v1/transaction_filter.proto\x12\x16\x63om.daml.ledger.api.v1\x1a\"com/daml/ledger/api/v1/value.proto\"\xc4\x01\n\x11TransactionFilter\x12W\n\x10\x66ilters_by_party\x18\x01 \x03(\x0b\x32=.com.daml.ledger.api.v1.TransactionFilter.FiltersByPartyEntry\x1aV\n\x13\x46iltersByPartyEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12.\n\x05value\x18\x02 \x01(\x0b\x32\x1f.com.daml.ledger.api.v1.Filters:\x02\x38\x01\"F\n\x07\x46ilters\x12;\n\tinclusive\x18\x01 \x01(\x0b\x32(.com.daml.ledger.api.v1.InclusiveFilters\"L\n\x10InclusiveFilters\x12\x38\n\x0ctemplate_ids\x18\x01 \x03(\x0b\x32\".com.daml.ledger.api.v1.IdentifierBN\n\x16\x63om.daml.ledger.api.v1B\x1bTransactionFilterOuterClass\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3'
  ,
  dependencies=[com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2.DESCRIPTOR,])




_TRANSACTIONFILTER_FILTERSBYPARTYENTRY = _descriptor.Descriptor(
  name='FiltersByPartyEntry',
  full_name='com.daml.ledger.api.v1.TransactionFilter.FiltersByPartyEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='com.daml.ledger.api.v1.TransactionFilter.FiltersByPartyEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='com.daml.ledger.api.v1.TransactionFilter.FiltersByPartyEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=222,
  serialized_end=308,
)

_TRANSACTIONFILTER = _descriptor.Descriptor(
  name='TransactionFilter',
  full_name='com.daml.ledger.api.v1.TransactionFilter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='filters_by_party', full_name='com.daml.ledger.api.v1.TransactionFilter.filters_by_party', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTIONFILTER_FILTERSBYPARTYENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=112,
  serialized_end=308,
)


_FILTERS = _descriptor.Descriptor(
  name='Filters',
  full_name='com.daml.ledger.api.v1.Filters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='inclusive', full_name='com.daml.ledger.api.v1.Filters.inclusive', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=310,
  serialized_end=380,
)


_INCLUSIVEFILTERS = _descriptor.Descriptor(
  name='InclusiveFilters',
  full_name='com.daml.ledger.api.v1.InclusiveFilters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='template_ids', full_name='com.daml.ledger.api.v1.InclusiveFilters.template_ids', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=382,
  serialized_end=458,
)

_TRANSACTIONFILTER_FILTERSBYPARTYENTRY.fields_by_name['value'].message_type = _FILTERS
_TRANSACTIONFILTER_FILTERSBYPARTYENTRY.containing_type = _TRANSACTIONFILTER
_TRANSACTIONFILTER.fields_by_name['filters_by_party'].message_type = _TRANSACTIONFILTER_FILTERSBYPARTYENTRY
_FILTERS.fields_by_name['inclusive'].message_type = _INCLUSIVEFILTERS
_INCLUSIVEFILTERS.fields_by_name['template_ids'].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2._IDENTIFIER
DESCRIPTOR.message_types_by_name['TransactionFilter'] = _TRANSACTIONFILTER
DESCRIPTOR.message_types_by_name['Filters'] = _FILTERS
DESCRIPTOR.message_types_by_name['InclusiveFilters'] = _INCLUSIVEFILTERS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TransactionFilter = _reflection.GeneratedProtocolMessageType('TransactionFilter', (_message.Message,), {

  'FiltersByPartyEntry' : _reflection.GeneratedProtocolMessageType('FiltersByPartyEntry', (_message.Message,), {
    'DESCRIPTOR' : _TRANSACTIONFILTER_FILTERSBYPARTYENTRY,
    '__module__' : 'com.daml.ledger.api.v1.transaction_filter_pb2'
    # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.TransactionFilter.FiltersByPartyEntry)
    })
  ,
  'DESCRIPTOR' : _TRANSACTIONFILTER,
  '__module__' : 'com.daml.ledger.api.v1.transaction_filter_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.TransactionFilter)
  })
_sym_db.RegisterMessage(TransactionFilter)
_sym_db.RegisterMessage(TransactionFilter.FiltersByPartyEntry)

Filters = _reflection.GeneratedProtocolMessageType('Filters', (_message.Message,), {
  'DESCRIPTOR' : _FILTERS,
  '__module__' : 'com.daml.ledger.api.v1.transaction_filter_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Filters)
  })
_sym_db.RegisterMessage(Filters)

InclusiveFilters = _reflection.GeneratedProtocolMessageType('InclusiveFilters', (_message.Message,), {
  'DESCRIPTOR' : _INCLUSIVEFILTERS,
  '__module__' : 'com.daml.ledger.api.v1.transaction_filter_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.InclusiveFilters)
  })
_sym_db.RegisterMessage(InclusiveFilters)


DESCRIPTOR._options = None
_TRANSACTIONFILTER_FILTERSBYPARTYENTRY._options = None
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Write-Side types
----------------

The :mod:`dazl.model.writing` module contains the Python classes used on the write-side of the
Ledger API.

.. autoclass:: Command
   :members:

.. autoclass:: CreateCommand
   :members:

.. autoclass:: ExerciseCommand
   :members:
"""
from dataclasses import dataclass, fields
from datetime import timedelta
from typing import TYPE_CHECKING, Any, Collection, List, Mapping, Optional, Sequence, Union
import uuid
import warnings

from ..damlast.daml_lf_1 import TypeConName
from ..damlast.daml_types import con
from ..damlast.lookup import find_choice
from ..damlast.protocols import SymbolLookup
from ..prim import ContractId, Party
from ..util.typing import safe_cast

if TYPE_CHECKING:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        from .types import Type, TypeReference

    from ..values import Context, ValueMapper

CommandsOrCommandSequence = Union[None, "Command", Sequence[Optional["Command"]]]
EventHandlerResponse = Union[CommandsOrCommandSequence, "CommandBuilder", "CommandPayload"]


class Command:
    """
    Base class for write-side commands.
    """


@dataclass(init=False, frozen=True)
class CreateCommand(Command):
    """
    A command that creates a contract without any predecessors.

    .. attribute:: CreateCommand.template

        Refers to the type of a template. This can be passed in as a ``str`` to the constructor,
        where it assumed to represent the ID or name of a template.

    .. attribute:: CreateCommand.arguments

        The arguments to the create (as a ``dict``).
    """

    __slots__ = ("template_type", "arguments", "_template_type_deprecated")

    template_type: "TypeConName"
    arguments: "Mapping[str, Any]"
    _template_type_deprecated: "TypeReference"

    def __init__(self, template: "Union[str, TypeConName, Type]", arguments=None):
        from ..damlast.compat import parse_template

        template_type, template_type_deprecated = parse_template(template)
        object.__setattr__(self, "template_type", template_type)
        object.__setattr__(self, "arguments", arguments or dict())
        object.__setattr__(self, "_template_type_deprecated", template_type_deprecated)

    @property
    def template(self) -> "TypeReference":
        warnings.warn(
            "CreateCommand.template is deprecated; use CreateCommand.template_type instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._template_type_deprecated

    def replace(self, template: "Union[None, str, Type]" = None, arguments=None):
        """
        Create a new :class:`CreateCommand` with the same identifier as this command, but with new
        values for its parameters.

        :param template:
            The new value of the `template` field, or `None` to reuse the existing value.
        :param arguments:
            The new value of the `arguments` field, or `None` to reuse the existing value.
        """
        warnings.warn(
            "CreateCommand.replace is deprecated; simply construct a CreateCommand with the "
            "desired values instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)

            from .types import Type, UnresolvedTypeReference

            if template is not None:
                template = (
                    template if isinstance(template, Type) else UnresolvedTypeReference(template)
                )
            return CreateCommand(
                template if template is not None else self.template_type,
                arguments if arguments is not None else self.arguments,
            )

    def __repr__(self):
        return f"<create {self.template_type} {self.arguments}>"


@dataclass(init=False, frozen=True)
class ExerciseCommand(Command):
    """
    A command that exercises a choice on a pre-existing contract.

    .. attribute:: ExerciseCommand.contract

        The :class:`ContractId` on which a choice is being exercised.

    .. attribute:: ExerciseCommand.choice

        Refers to a choice (either a :class:`ChoiceRef` or a :class:`ChoiceMetadata`).
        This can be passed in as a ``str`` to the constructor, where it assumed to represent the
        name of a choice.

    .. attribute:: ExerciseCommand.arguments

        The arguments to the exercise choice (as a ``dict``).

    Note that when an ``ExerciseCommand`` is created, an additional ``template_id`` parameter can
    be supplied to the constructor to aid in disambiguation of the specific choice being invoked.
    In some situations involving composite commands, a ``template_id`` must eventually be supplied
    before a choice can be exercised. If this ``template_id`` is specified, the ``contract`` and
    ``choice`` are both tagged with this ID.

    Instance methods:

    .. automethod:: replace
    """

    __slots__ = ("contract", "choice", "arguments")

    contract: "ContractId"
    choice: str
    arguments: "Optional[Any]"

    def __init__(self, contract: "ContractId", choice: str, arguments: "Optional[Any]" = None):
        object.__setattr__(self, "contract", safe_cast(ContractId, contract))
        object.__setattr__(self, "choice", safe_cast(str, choice))
        object.__setattr__(self, "arguments", dict(arguments) if arguments is not None else dict())

    def replace(self, contract=None, choice=None, arguments=None):
        """
        Create a new :class:`ExerciseCommand` with the same identifier as this command, but with new
        values for its parameters.

        :param contract:
            The new value of the `contract` field, or `None` to reuse the existing value.
            The same type coercion rules used in the constructor apply here.
        :param choice:
            The new value of the `choice` field, or `None` to reuse the existing value.
            The same type coercion rules used in the constructor apply here.
        :param arguments:
            The new value of the `choice` field, or `None` to reuse the existing value.
        """
        warnings.warn(
            "ExerciseCommand.replace is deprecated; simply construct a ExerciseCommand with the "
            "desired values instead.",
            DeprecationWarning,
        )
        return ExerciseCommand(
            contract if contract is not None else self.contract,
            choice if choice is not None else self.choice,
            arguments if arguments is not None else self.arguments,
        )

    def __repr__(self):
        return f"<exercise '{self.contract.value}' {self.choice} with {self.arguments}>"


@dataclass(init=False, frozen=True)
class ExerciseByKeyCommand(Command):
    template_type: "TypeConName"
    contract_key: "Any"
    choice: str
    choice_argument: "Any"
    _template_type_deprecated: "TypeReference"

    def __init__(
        self,
        template: "Union[str, TypeConName, TypeReference]",
        contract_key: "Any",
        choice: str,
        choice_argument: "Any",
    ):
        from ..damlast.compat import parse_template

        template_type, template_type_deprecated = parse_template(template)
        object.__setattr__(self, "template_type", template_type)
        object.__setattr__(self, "contract_key", contract_key)
        object.__setattr__(self, "choice", choice)
        object.__setattr__(self, "choice_argument", choice_argument)
        object.__setattr__(self, "_template_type_deprecated", template_type_deprecated)

    @property
    def template(self) -> "TypeReference":
        warnings.warn(
            "ExerciseByKeyCommand.template is deprecated; use ExerciseByKeyCommand.template_type "
            "instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._template_type_deprecated


@dataclass(init=False, frozen=True)
class CreateAndExerciseCommand(Command):
    template_type: "TypeConName"
    arguments: "Mapping[str, Any]"
    choice: str
    choice_argument: "Any"
    _template_type_deprecated: "TypeReference"

    def __init__(
        self,
        template: "Union[str, TypeConName, TypeReference]",
        arguments: "Mapping[str, Any]",
        choice: str,
        choice_argument: "Any",
    ):
        from ..damlast.compat import parse_template

        template_type, template_type_deprecated = parse_template(template)
        object.__setattr__(self, "template_type", template_type)
        object.__setattr__(self, "arguments", arguments)
        object.__setattr__(self, "choice", choice)
        object.__setattr__(self, "choice_argument", choice_argument)
        object.__setattr__(self, "_template_type_deprecated", template_type_deprecated)

    @property
    def template(self) -> "TypeReference":
        warnings.warn(
            "CreateAndExerciseCommand.template is deprecated; use "
            "CreateAndExerciseCommand.template_type instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._template_type_deprecated


class CommandBuilder:
    """
    Builder class for generating commands to be sent to the ledger.
    """

    @classmethod
    def coerce(cls, obj, atomic_default=False) -> "CommandBuilder":
        """
        Create a :class:`CommandBuilder` from the objects that an event handler is allowed to
        return.

        :param obj:
        :param atomic_default:
        :return:
        """
        if isinstance(obj, CommandBuilder):
            return obj

        builder = CommandBuilder(atomic_default=atomic_default)
        if obj is not None:
            builder.append(obj)
        return builder

    def __init__(self, atomic_default=False):
        self._atomic_default = atomic_default
        self._commands = [[]]  # type: List[List[Command]]
        self._defaults = CommandDefaults()

    def defaults(
        self,
        party: Optional[Party] = None,
        ledger_id: Optional[str] = None,
        workflow_id: Optional[str] = None,
        application_id: Optional[str] = None,
        command_id: Optional[str] = None,
        deduplication_time: Optional[timedelta] = None,
    ) -> None:
        if party is not None:
            self._defaults.default_party = party
        if ledger_id is not None:
            self._defaults.ledger_id = ledger_id
        if workflow_id is not None:
            self._defaults.default_workflow_id = workflow_id
        if application_id is not None:
            self._defaults.default_application_id = application_id
        if command_id is not None:
            self._defaults.default_command_id = command_id
        if deduplication_time is not None:
            self._defaults.default_deduplication_time = deduplication_time

    def create(self, template, arguments=None) -> "CommandBuilder":
        return self.append(create(template, arguments=arguments))

    def exercise(self, contract, choice, arguments=None) -> "CommandBuilder":
        return self.append(exercise(contract, choice, arguments=arguments))

    def create_and_exercise(
        self, template, create_arguments, choice_name, choice_arguments=None
    ) -> "CommandBuilder":
        return self.append(
            create_and_exercise(template, create_arguments, choice_name, choice_arguments)
        )

    def append(self, *commands: CommandsOrCommandSequence) -> "CommandBuilder":
        """
        Append one or more commands, or list of commands to the :class:`CommandBuilder` in flight.
        This method respects the value of ``atomic_default`` that this object was constructed with.
        In order to force commands to be submitted either atomically, use :meth:`append_atomically`.
        To allow these commands to be submitted in parallel use :meth:`append_nonatomically`.

        :param commands: One or more commands, or list of commands to be submitted to the ledger.
        :return: This object.
        """
        if self._atomic_default:
            # a command builder that defaults to being atomic will put all commands in a single
            # transaction; build on the very first transaction
            self._commands[0].extend(flatten_command_sequence(commands))
            return self
        else:
            return self.append_nonatomically(*commands)

    def append_atomically(self, *commands: Union[Command, Sequence[Command]]) -> "CommandBuilder":
        self._commands.extend([flatten_command_sequence(commands)])
        return self

    def append_nonatomically(
        self, *commands: Union[Command, Sequence[Command]]
    ) -> "CommandBuilder":
        self._commands.extend([[cmd] for cmd in flatten_command_sequence(commands)])
        return self

    def build(self, defaults: "Optional[CommandDefaults]" = None) -> "Collection[CommandPayload]":
        """
        Return a collection of commands.
        """
        if defaults is None:
            raise ValueError("defaults must currently be specified")

        command_id = (
            defaults.default_command_id or self._defaults.default_command_id or uuid.uuid4().hex
        )

        return [
            CommandPayload(
                party=defaults.default_party or self._defaults.default_party,
                ledger_id=defaults.default_ledger_id or self._defaults.default_ledger_id,
                workflow_id=defaults.default_workflow_id or self._defaults.default_workflow_id,
                application_id=defaults.default_application_id
                or self._defaults.default_application_id,
                command_id=command_id,
                deduplication_time=defaults.default_deduplication_time
                or self._defaults.default_deduplication_time,
                commands=commands,
            )
            for i, commands in enumerate(self._commands)
            if commands
        ]

    def __format__(self, format_spec):
        if format_spec == "c":
            return str(self._commands)
        else:
            return repr(self)

    def __repr__(self):
        return f"CommandBuilder({self._commands})"


def flatten_command_sequence(commands: Sequence[CommandsOrCommandSequence]) -> List[Command]:
    """
    Convert a list of mixed commands, ``None``, and list of commands into an ordered sequence of
    non-``None`` commands.
    """
    ret = []  # type: List[Command]
    errors = []

    for i, obj in enumerate(commands):
        if obj is not None:
            if isinstance(obj, Command):
                ret.append(obj)
            else:
                try:
                    cmd_iter = iter(obj)
                except TypeError:
                    errors.append(((i,), obj))
                    continue
                for j, cmd in enumerate(cmd_iter):
                    if isinstance(cmd, Command):
                        ret.append(cmd)
                    else:
                        errors.append(((i, j), cmd))
    if errors:
        raise ValueError(
            f"Failed to interpret some elements as Commands in the list: " f"$[{index}] = {command}"
            for index, command in errors
        )
    return ret


@dataclass
class CommandDefaults:
    """
    Values to use for a :class:`Command` when no value is specified with the creation of the
    command.
    """

    default_party: Optional[Party] = None
    default_ledger_id: Optional[str] = None
    default_workflow_id: Optional[str] = None
    default_application_id: Optional[str] = None
    default_command_id: Optional[str] = None
    default_deduplication_time: Optional[timedelta] = None


@dataclass(frozen=True)
class CommandPayload:
    """
    A request to mutate active state of the ledger.

    .. attribute:: CommandPayload.party
        The party submitting the request.
    .. attribute:: CommandPayload.application_id:
        An optional application ID to accompany the request.
    .. attribute:: CommandPayload.command_id:
        A hash that represents the BIM commitment.
    .. attribute:: CommandPayload.deduplication_time:
        The maximum time interval before the client should consider this command expired.
    .. attribute:: CommandPayload.commands
        A sequence of commands to submit to the ledger. These commands are submitted atomically
        (in other words, they all succeed or they all fail).
    .. attribute:: CommandPayload.deduplication_time:
        The length of the time window during which all commands with the same party and command ID
        will be deduplicated. Duplicate commands submitted before the end of this window return an
        ``ALREADY_EXISTS`` error.
    """

    party: Party
    ledger_id: str
    workflow_id: str
    application_id: str
    command_id: str
    commands: "Sequence[Command]"
    deduplication_time: "Optional[timedelta]" = None

    def __post_init__(self):
        missing_fields = [
            field.name
            for field in fields(self)
            if field.name != "deduplication_time" and getattr(self, field.name) is None
        ]
        if missing_fields:
            raise ValueError(
                f"Some fields are set to None when they are required: " f"{missing_fields}"
            )


def create(template, arguments=None):
    if not isinstance(template, str):
        raise ValueError(
            "template must be a string name, a template type, or an instantiated template"
        )

    return CreateCommand(template, arguments)


def exercise(contract, choice, arguments=None):
    if not isinstance(choice, str):
        raise ValueError(
            "choice must be a string name, a template type, " "or an instantiated template"
        )

    return ExerciseCommand(contract, choice, arguments)


def exercise_by_key(template, contract_key, choice_name, choice_argument):
    return ExerciseByKeyCommand(template, contract_key, choice_name, choice_argument)


def create_and_exercise(template, create_arguments, choice_name, choice_argument):
    return CreateAndExerciseCommand(template, create_arguments, choice_name, choice_argument)


class Serializer:
    """
    Serializer interface for objects on the write-side of the API.
    """

    def serialize_value(self, tt: "Type", obj: "Any") -> "Any":
        raise NotImplementedError("serialize_value requires an implementation")

    def serialize_command(self, command: "Any") -> "Any":
        raise NotImplementedError("serialize_command requires an implementation")


class AbstractSerializer(Serializer):
    """
    Implementation of :class:`Serializer` that helps enforce that all possible cases of type
    serialization have been implemented.
    """

    def __init__(self, lookup: "SymbolLookup"):
        self.lookup = lookup

    @property
    def mapper(self) -> "ValueMapper":
        raise NotImplementedError(f"{type(self)}.mapper() must be defined")

    def serialize_value(self, tt: "Type", obj: Any) -> "Any":
        from ..values import Context

        return Context(self.mapper, self.lookup).convert(tt, obj)

    def serialize_commands(self, commands: "Sequence[Command]") -> "Sequence[Any]":
        return [self.serialize_command(cmd) for cmd in commands]

    def serialize_command(self, command: "Command") -> "Any":
        if isinstance(command, CreateCommand):
            name = self.lookup.template_name(command.template_type)
            value = self.serialize_value(con(name), command.arguments)
            return self.serialize_create_command(name, value)

        elif isinstance(command, ExerciseCommand):
            template = self.lookup.template(command.contract.value_type)
            choice = find_choice(template, command.choice)
            args = self.serialize_value(choice.arg_binder.type, command.arguments)
            return self.serialize_exercise_command(command.contract, choice.name, args)

        elif isinstance(command, CreateAndExerciseCommand):
            name = self.lookup.template_name(command.template_type)
            template = self.lookup.template(name)
            create_value = self.serialize_value(con(name), command.arguments)
            choice = find_choice(template, command.choice)
            choice_args = self.serialize_value(choice.arg_binder.type, command.choice_argument)
            return self.serialize_create_and_exercise_command(
                name, create_value, choice.name, choice_args
            )

        elif isinstance(command, ExerciseByKeyCommand):
            name = self.lookup.template_name(command.template_type)
            template = self.lookup.template(name)
            key_value = self.serialize_value(template.key.type, command.contract_key)
            choice = find_choice(template, command.choice)
            choice_args = self.serialize_value(choice.arg_binder.type, command.choice_argument)
            return self.serialize_exercise_by_key_command(name, key_value, choice.name, choice_args)

        else:
            raise ValueError(f"unknown Command type: {command!r}")

    def serialize_create_command(self, name: "TypeConName", template_args: "Any") -> "Any":
        raise NotImplementedError("serialize_create_command requires an implementation")

    def serialize_exercise_command(
        self, contract_id: "ContractId", choice_name: str, choice_args: "Any"
    ) -> "Any":
        raise NotImplementedError("serialize_exercise_command requires an implementation")

    def serialize_exercise_by_key_command(
        self, name: "TypeConName", key_arguments: "Any", choice_name: str, choice_arguments: "Any"
    ) -> "Any":
        raise NotImplementedError("serialize_exercise_by_key_command requires an implementation")

    def serialize_create_and_exercise_command(
        self, name: "TypeConName", create_args: "Any", choice_name: str, choice_arguments: "Any"
    ) -> "Any":
        raise NotImplementedError(
            "serialize_create_and_exercise_command requires an implementation"
        )



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/command_service.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import commands_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_commands__pb2
from . import trace_context_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2
from . import transaction_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/daml/ledger/api/v1/command_service.proto',
  package='com.daml.ledger.api.v1',
  syntax='proto3',
  serialized_options=b'\n\026com.daml.ledger.api.v1B\030CommandServiceOuterClass\252\002\026Com.Daml.Ledger.Api.V1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n,com/daml/ledger/api/v1/command_service.proto\x12\x16\x63om.daml.ledger.api.v1\x1a%com/daml/ledger/api/v1/commands.proto\x1a*com/daml/ledger/api/v1/trace_context.proto\x1a(com/daml/ledger/api/v1/transaction.proto\x1a\x1bgoogle/protobuf/empty.proto\"\x88\x01\n\x14SubmitAndWaitRequest\x12\x32\n\x08\x63ommands\x18\x01 \x01(\x0b\x32 .com.daml.ledger.api.v1.Commands\x12<\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32$.com.daml.ledger.api.v1.TraceContext\"?\n%SubmitAndWaitForTransactionIdResponse\x12\x16\n\x0etransaction_id\x18\x01 \x01(\t\"_\n#SubmitAndWaitForTransactionResponse\x12\x38\n\x0btransaction\x18\x01 \x01(\x0b\x32#.com.daml.ledger.api.v1.Transaction\"g\n\'SubmitAndWaitForTransactionTreeResponse\x12<\n\x0btransaction\x18\x01 \x01(\x0b\x32\'.com.daml.ledger.api.v1.TransactionTree2\x94\x04\n\x0e\x43ommandService\x12U\n\rSubmitAndWait\x12,.com.daml.ledger.api.v1.SubmitAndWaitRequest\x1a\x16.google.protobuf.Empty\x12\x8c\x01\n\x1dSubmitAndWaitForTransactionId\x12,.com.daml.ledger.api.v1.SubmitAndWaitRequest\x1a=.com.daml.ledger.api.v1.SubmitAndWaitForTransactionIdResponse\x12\x88\x01\n\x1bSubmitAndWaitForTransaction\x12,.com.daml.ledger.api.v1.SubmitAndWaitRequest\x1a;.com.daml.ledger.api.v1.SubmitAndWaitForTransactionResponse\x12\x90\x01\n\x1fSubmitAndWaitForTransactionTree\x12,.com.daml.ledger.api.v1.SubmitAndWaitRequest\x1a?.com.daml.ledger.api.v1.SubmitAndWaitForTransactionTreeResponseBK\n\x16\x63om.daml.ledger.api.v1B\x18\x43ommandServiceOuterClass\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3'
  ,
  dependencies=[com_dot_daml_dot_ledger_dot_api_dot_v1_dot_commands__pb2.DESCRIPTOR,com_dot_daml_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2.DESCRIPTOR,com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__pb2.DESCRIPTOR,google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,])




_SUBMITANDWAITREQUEST = _descriptor.Descriptor(
  name='SubmitAndWaitRequest',
  full_name='com.daml.ledger.api.v1.SubmitAndWaitRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='commands', full_name='com.daml.ledger.api.v1.SubmitAndWaitRequest.commands', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.daml.ledger.api.v1.SubmitAndWaitRequest.trace_context', index=1,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=227,
  serialized_end=363,
)


_SUBMITANDWAITFORTRANSACTIONIDRESPONSE = _descriptor.Descriptor(
  name='SubmitAndWaitForTransactionIdResponse',
  full_name='com.daml.ledger.api.v1.SubmitAndWaitForTransactionIdResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transaction_id', full_name='com.daml.ledger.api.v1.SubmitAndWaitForTransactionIdResponse.transaction_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=365,
  serialized_end=428,
)


_SUBMITANDWAITFORTRANSACTIONRESPONSE = _descriptor.Descriptor(
  name='SubmitAndWaitForTransactionResponse',
  full_name='com.daml.ledger.api.v1.SubmitAndWaitForTransactionResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transaction', full_name='com.daml.ledger.api.v1.SubmitAndWaitForTransactionResponse.transaction', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=430,
  serialized_end=525,
)


_SUBMITANDWAITFORTRANSACTIONTREERESPONSE = _descriptor.Descriptor(
  name='SubmitAndWaitForTransactionTreeResponse',
  full_name='com.daml.ledger.api.v1.SubmitAndWaitForTransactionTreeResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transaction', full_name='com.daml.ledger.api.v1.SubmitAndWaitForTransactionTreeResponse.transaction', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=527,
  serialized_end=630,
)

_SUBMITANDWAITREQUEST.fields_by_name['commands'].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_commands__pb2._COMMANDS
_SUBMITANDWAITREQUEST.fields_by_name['trace_context'].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
_SUBMITANDWAITFORTRANSACTIONRESPONSE.fields_by_name['transaction'].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__pb2._TRANSACTION
_SUBMITANDWAITFORTRANSACTIONTREERESPONSE.fields_by_name['transaction'].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__pb2._TRANSACTIONTREE
DESCRIPTOR.message_types_by_name['SubmitAndWaitRequest'] = _SUBMITANDWAITREQUEST
DESCRIPTOR.message_types_by_name['SubmitAndWaitForTransactionIdResponse'] = _SUBMITANDWAITFORTRANSACTIONIDRESPONSE
DESCRIPTOR.message_types_by_name['SubmitAndWaitForTransactionResponse'] = _SUBMITANDWAITFORTRANSACTIONRESPONSE
DESCRIPTOR.message_types_by_name['SubmitAndWaitForTransactionTreeResponse'] = _SUBMITANDWAITFORTRANSACTIONTREERESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SubmitAndWaitRequest = _reflection.GeneratedProtocolMessageType('SubmitAndWaitRequest', (_message.Message,), {
  'DESCRIPTOR' : _SUBMITANDWAITREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.command_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.SubmitAndWaitRequest)
  })
_sym_db.RegisterMessage(SubmitAndWaitRequest)

SubmitAndWaitForTransactionIdResponse = _reflection.GeneratedProtocolMessageType('SubmitAndWaitForTransactionIdResponse', (_message.Message,), {
  'DESCRIPTOR' : _SUBMITANDWAITFORTRANSACTIONIDRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.command_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.SubmitAndWaitForTransactionIdResponse)
  })
_sym_db.RegisterMessage(SubmitAndWaitForTransactionIdResponse)

SubmitAndWaitForTransactionResponse = _reflection.GeneratedProtocolMessageType('SubmitAndWaitForTransactionResponse', (_message.Message,), {
  'DESCRIPTOR' : _SUBMITANDWAITFORTRANSACTIONRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.command_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.SubmitAndWaitForTransactionResponse)
  })
_sym_db.RegisterMessage(SubmitAndWaitForTransactionResponse)

SubmitAndWaitForTransactionTreeResponse = _reflection.GeneratedProtocolMessageType('SubmitAndWaitForTransactionTreeResponse', (_message.Message,), {
  'DESCRIPTOR' : _SUBMITANDWAITFORTRANSACTIONTREERESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.command_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.SubmitAndWaitForTransactionTreeResponse)
  })
_sym_db.RegisterMessage(SubmitAndWaitForTransactionTreeResponse)


DESCRIPTOR._options = None

_COMMANDSERVICE = _descriptor.ServiceDescriptor(
  name='CommandService',
  full_name='com.daml.ledger.api.v1.CommandService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=633,
  serialized_end=1165,
  methods=[
  _descriptor.MethodDescriptor(
    name='SubmitAndWait',
    full_name='com.daml.ledger.api.v1.CommandService.SubmitAndWait',
    index=0,
    containing_service=None,
    input_type=_SUBMITANDWAITREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SubmitAndWaitForTransactionId',
    full_name='com.daml.ledger.api.v1.CommandService.SubmitAndWaitForTransactionId',
    index=1,
    containing_service=None,
    input_type=_SUBMITANDWAITREQUEST,
    output_type=_SUBMITANDWAITFORTRANSACTIONIDRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SubmitAndWaitForTransaction',
    full_name='com.daml.ledger.api.v1.CommandService.SubmitAndWaitForTransaction',
    index=2,
    containing_service=None,
    input_type=_SUBMITANDWAITREQUEST,
    output_type=_SUBMITANDWAITFORTRANSACTIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SubmitAndWaitForTransactionTree',
    full_name='com.daml.ledger.api.v1.CommandService.SubmitAndWaitForTransactionTree',
    index=3,
    containing_service=None,
    input_type=_SUBMITANDWAITREQUEST,
    output_type=_SUBMITANDWAITFORTRANSACTIONTREERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_COMMANDSERVICE)

DESCRIPTOR.services_by_name['CommandService'] = _COMMANDSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from asyncio import Future, ensure_future, gather, get_event_loop, wait
from concurrent.futures import ALL_COMPLETED
from dataclasses import asdict, dataclass, field, fields, replace
from datetime import datetime
from io import StringIO
import reprlib
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Collection,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
)
import uuid

from .. import LOG
from ..damlast.daml_lf_1 import TypeConName
from ..model.core import (
    ContractContextualData,
    ContractContextualDataCollection,
    ContractId,
    ContractMatch,
    ContractsState,
    Party,
)
from ..model.ledger import LedgerMetadata
from ..model.network import OAuthSettings, connection_settings
from ..model.reading import (
    ActiveContractSetEvent,
    BaseEvent,
    ContractArchiveEvent,
    ContractCreateEvent,
    ContractFilter,
    InitEvent,
    OffsetEvent,
    PackagesAddedEvent,
    ReadyEvent,
    TransactionEndEvent,
    TransactionFilter,
    TransactionStartEvent,
)
from ..model.writing import CommandBuilder, CommandDefaults, CommandPayload, EventHandlerResponse
from ..prim import TimeDeltaLike, to_timedelta
from ..protocols import LedgerClient, LedgerNetwork
from ..util.asyncio_util import ServiceQueue, completed, named_gather
from ..util.prim_natural import n_things
from ..util.typing import safe_cast
from ._writer_verify import ValidateSerializer
from .bots import Bot, BotCallback, BotCollection, BotFilter
from .config import NetworkConfig, PartyConfig
from .state import ActiveContractSet

if TYPE_CHECKING:
    from ._network_client_impl import _NetworkImpl
    from .api import PartyClient


T = TypeVar("T")


class _PartyClientImpl:
    def __init__(self, parent: "_NetworkImpl", party: "Party"):
        self.parent = parent
        self.metrics = parent._metrics
        self.invoker = parent.invoker
        self.party = party

        self._config_values = dict()
        self._config = None  # type: Optional[NetworkConfig]
        self._pool = None  # type: Optional[LedgerNetwork]
        self._pool_fut = None  # type: Optional[Awaitable[LedgerNetwork]]
        self._client_fut = None  # type: Optional[Awaitable[LedgerClient]]
        self._ready_fut = self.invoker.create_future()
        self._known_packages = set()  # type: Set[str]

        self._acs = ActiveContractSet(self.invoker, self.parent.lookup)
        self.bots = BotCollection(party)
        self._reader = _PartyClientReaderState()
        self._writer = _PartyClientWriterState()

    def connect_in(self, pool: "LedgerNetwork") -> "Future":
        self._config = config = self.resolved_config()

        oauth = None
        if config.oauth_token or config.oauth_client_id:
            oauth = OAuthSettings(
                token=config.oauth_token,
                token_uri=config.oauth_token_uri,
                refresh_token=config.oauth_refresh_token,
                client_id=config.oauth_client_id,
                client_secret=config.oauth_client_secret,
                id_token=config.oauth_id_token,
                redirect_uri=config.oauth_redirect_uri,
                auth_url=config.oauth_auth_url,
                auth_audience=config.oauth_audience,
                auth_ca_file=config.oauth_ca_file,
            )

        settings, url_prefix = connection_settings(
            config.url,
            self.party,
            verify_ssl=config.verify_ssl,
            ca_file=config.ca_file,
            cert_file=config.cert_file,
            cert_key_file=config.cert_key_file,
            oauth=oauth,
        )

        self._pool = pool
        self._client_fut = ensure_future(pool.connect(self.party, settings, url_prefix))
        return self._client_fut

    def set_config(self, **kwargs):
        self._config_values.update(kwargs)

    def resolved_config(self) -> "PartyConfig":
        parent_base_config = asdict(self.parent.resolved_anonymous_config())
        parent_party_config = (
            _find_party_config(self.parent.get_config_raw("parties", ()), self.party) or {}
        )

        config_names = {f.name for f in fields(PartyConfig)}
        all_config = dict()
        all_config.update(parent_base_config)
        all_config.update(parent_party_config)
        all_config.update(self._config_values)
        return PartyConfig(**{k: v for k, v in all_config.items() if k in config_names})

    def initialize(self, current_time: datetime, metadata: LedgerMetadata) -> Awaitable[None]:
        """
        Initialize the state of the ledger.

        :param current_time:
            The ledger time to publish as part of the :class:`InitEvent`.
        :param metadata:
            Information about the connected ledger.
        """
        evt = InitEvent(
            self, self.party, current_time, metadata.ledger_id, self.parent.lookup, metadata._store
        )
        return self.emit_event(evt)

    def ready(self) -> Awaitable[None]:
        return self._ready_fut

    # region Event Handler Management

    # noinspection PyShadowingBuiltins
    def add_event_handler(
        self,
        key: str,
        handler: "BotCallback",
        filter: "Optional[BotFilter]",
        party_client: "PartyClient",
    ) -> "Bot":
        if party_client is None:
            raise ValueError("a party_client is required here")
        return self.bots.add_single(key, handler, filter, party_client=party_client)

    def emit_event(self, data: BaseEvent) -> "Awaitable[Any]":
        """
        Emit an event.

        :param data:
            The event to raise.
        :return:
            An Awaitable that is resolved when the commands that resulted from event callbacks
            have been completed either successfully or unsuccessfully.
        """
        return self.bots.notify(data)

    async def emit_ready(self, metadata: LedgerMetadata, time: datetime, offset: str) -> None:
        """
        Emit a ready event specific to this client. This may also emit initial create events and
        initial package added events.

        :param metadata:
            The :class:`LedgerMetadata` at the network level.
        :param time:
            The current time.
        :param offset:
            The ledger offset.
        :return:
            Nothing. This method "returns" when all events have been raised and their follow-ups
            have been processed.
        """
        ready_event = ReadyEvent(
            self, self.party, time, metadata.ledger_id, self.parent.lookup, metadata._store, offset
        )
        self._known_packages.update(self.parent.lookup.package_ids())
        await self.emit_event(ready_event)

        pkg_event = PackagesAddedEvent(
            self, self.party, time, metadata.ledger_id, self.parent.lookup, metadata._store, True
        )
        await self.emit_event(pkg_event)

    # endregion

    # region Active/Historical Contract Set management

    def find_by_id(self, cid: ContractId) -> ContractContextualData:
        return self._acs.get(cid)

    def find(
        self, template: Any, match: ContractMatch = None, include_archived: bool = False
    ) -> ContractContextualDataCollection:
        return self._acs.read_full(template, match, include_archived=include_archived)

    def find_active(self, template: Any, match: ContractMatch = None) -> ContractsState:
        return self._acs.read_active(template, match)

    def find_historical(
        self, template: Any, match: ContractMatch = None
    ) -> ContractContextualDataCollection:
        return self._acs.read_full(template, match, include_archived=True)

    def find_nonempty(
        self, template: Any, match: ContractMatch, min_count: int = 1, timeout: float = 30
    ):
        return self._acs.read_async(template, match, min_count=min_count)

    # endregion

    # region Read Path

    async def read_acs(
        self, until_offset: "Optional[str]", raise_events: bool
    ) -> "Tuple[Optional[str], Future]":
        """
        Initial bootstrap of events from the read side. Only one instance of this coroutine
        should be active at a time per client. An initial block of events is read using the
        Active Contract Set service, and the transaction stream is then subscribed to in order to
        ensure that this client is caught up to the specified offset.

        :param until_offset:
            The destination ledger offset to read up until to. If not included, then the client
            attempts to read as many transactions as is currently available.
        :param raise_events:
            ``True`` to raise transaction- and contract-related events to event handlers;
            ``False`` to suppress this behavior and only update local state within the client
            (offset information and active contract set).
        :return:
            The ledger offset that the reader ended at and a Future that is resolved when all event
            handlers' follow-ups have either successfully or unsuccessfully completed.

            If the PartyClient is _not_ configured to use the ActiveContractService, this function
            immediately returns `None`, and a resolved future.
        """
        if not self._config.use_acs_service:
            LOG.info(
                "read_acs(%r, %r) aborted for party %r because it was configured to not use "
                "the ACS service",
                until_offset,
                raise_events,
                self.party,
            )
            from asyncio import get_event_loop

            fut = get_event_loop().create_future()
            fut.set_result(None)
            return None, fut

        LOG.info("Calling read_acs(%r, %r) for party: %r", until_offset, raise_events, self.party)
        client = await self._client_fut

        contract_filter = ContractFilter(
            templates=await self._template_filter(), party_groups=self._config.party_groups
        )

        acs = await client.active_contracts(contract_filter)
        for acs_evt in acs:
            await self._process_transaction_stream_event(acs_evt, False)
        LOG.info("read_acs() finished at offset %r for party: %r", self._reader.offset, self.party)

        return await self.read_transactions(until_offset, raise_events)

    async def read_transactions(
        self, until_offset: "Optional[str]", raise_events: bool
    ) -> "Tuple[str, Future]":
        """
        Main processing method of events from the read side. Only one instance of this coroutine
        should be active at a time per client.

        :param until_offset:
            The destination ledger offset to read up until to. If not included, then the client
            attempts to read as many transactions as is currently available.
        :param raise_events:
            ``True`` to raise transaction- and contract-related events to event handlers;
            ``False`` to suppress this behavior and only update local state within the client
            (offset information and active contract set).
        :return:
            The ledger offset that the reader ended at and a Future that is resolved when all event
            handlers' follow-ups have either successfully or unsuccessfully completed.
        """
        LOG.verbose(
            "  read_transactions(%s, %s) with party %s | (groups %s), reader offset %s",
            until_offset,
            raise_events,
            self.party,
            self._config.party_groups,
            self._reader.offset,
        )
        client = await self._client_fut
        metadata = await self._pool.ledger()

        event_count = 0
        futs = []

        while (
            until_offset is None
            or self._reader.offset is None
            or self._reader.offset != until_offset
        ):
            LOG.verbose(
                "  read_transactions(%s, %s) with party %s | current offset: %s; "
                "destination offset: %s",
                until_offset,
                raise_events,
                self.party,
                self._reader.offset,
                until_offset,
            )

            # TODO: Find a more appropriate place to raise these events (or change the name of this
            #  method to make it clearer that it has a bigger mandate than simply transaction
            #  events)
            all_packages = set(self.parent.lookup.package_ids())
            new_package_ids = all_packages - self._known_packages
            if new_package_ids:
                pkg_event = PackagesAddedEvent(
                    self,
                    self.party,
                    None,
                    metadata.ledger_id,
                    self.parent.lookup,
                    metadata._store,
                    False,
                )
                self._known_packages.update(all_packages)
                await self.emit_event(pkg_event)

            # prepare a call to /events
            transaction_filter = TransactionFilter(
                current_offset=self._reader.offset,
                destination_offset=until_offset,
                templates=await self._template_filter(),
                party_groups=self._config.party_groups,
            )

            transaction_events = await client.events(transaction_filter)
            for event in transaction_events:
                futs.append(self._process_transaction_stream_event(event, raise_events))
            event_count += len(transaction_events)

            # if a destination offset is not specified, then only read one pass of transaction
            # events before exiting
            if until_offset is None:
                break

        self.metrics.party_offset(self.party, self._reader.offset)
        LOG.verbose(
            "  read_transactions(%s, %s) with party %s | reader offset %s",
            until_offset,
            raise_events,
            self.party,
            self._reader.offset,
        )
        futs = [fut for fut in futs if not fut.done()]
        if len(futs) == 0:
            return self._reader.offset, completed(None)
        elif len(futs) == 1:
            return self._reader.offset, futs[0]
        else:
            return self._reader.offset, gather(*futs, return_exceptions=True)

    async def read_end(self) -> str:
        client = await self._client_fut
        return await client.events_end()

    async def _template_filter(self) -> "Optional[Collection[TypeConName]]":
        if self._config.package_ids is not None:
            # first, ensure the packages we require are loaded
            metadata = await self._pool.ledger()
            await gather(*[metadata.package_loader.load(ref) for ref in self._config.package_ids])

            # now build a filter of the template names for all of the packages that we found
            names = [
                name
                for ref in self._config.package_ids
                for name in self.parent.lookup.template_names(f"{ref}:*")
            ]
            return names
        else:
            return None

    def _process_transaction_stream_event(self, event: Any, raise_events: bool) -> Future:
        """

        :param event:
        :param raise_events:
        """
        if isinstance(event, TransactionStartEvent):
            # Update the ACS before sharing this information with interested parties
            LOG.info(
                "Processing transaction: %s, %s, %s",
                event.offset,
                event.workflow_id,
                event.command_id,
            )
            for contract_event in event.contract_events:
                if isinstance(contract_event, ContractCreateEvent):
                    self._acs.handle_create(contract_event)
                elif isinstance(contract_event, ContractArchiveEvent):
                    self._acs.handle_archive(contract_event)

        elif isinstance(event, TransactionEndEvent):
            # Notify anything waiting on commands to complete that some of them will have
            # completed as a consequence of this transaction.
            for cmd in self._writer.inflight_commands:
                cmd.notify_read_done(event.command_id)

            LOG.debug(
                "evt recv: party %s, BIM %r (%s events)",
                self.party,
                event.command_id[0:7],
                len(event.contract_events),
            )

        elif isinstance(event, ActiveContractSetEvent):
            for contract_event in event.contract_events:
                self._acs.handle_create(contract_event)

        if raise_events:
            fut = self.emit_event(event)
        else:
            fut = completed(None)

        if isinstance(event, OffsetEvent):
            self._reader.offset = event.offset

        return fut

    # endregion

    # region Write path

    def write_commands(
        self,
        commands: EventHandlerResponse,
        ignore_errors: bool = False,
        workflow_id: "Optional[str]" = None,
        deduplication_time: "Optional[TimeDeltaLike]" = None,
    ) -> Awaitable[None]:
        """
        Submit a command or list of commands.

        :param commands:
            The commands to send to the server.
        :param ignore_errors:
            Whether errors should be ignored for purposes of terminating the client. If ``True``,
            then a failure to send this command does not necessarily end the client.
        :param workflow_id:
            The workflow ID to use to tag submitted commands.
        :param deduplication_time:
            The length of the time window during which all commands with the same party and command
            ID will be deduplicated. Duplicate commands submitted before the end of this window
            return an ``ALREADY_EXISTS`` error.
        :return:
            An ``asyncio.Future`` that is resolved right before the corresponding side effects have
            hit the event stream for this party. Synchronous errors are reported back immediately
            and not failed through this mechanism.
        """
        if workflow_id is None:
            workflow_id = uuid.uuid4().hex
        cb = CommandBuilder.coerce(commands, atomic_default=True)
        cb.defaults(
            workflow_id=workflow_id,
            deduplication_time=(
                to_timedelta(deduplication_time) if deduplication_time is not None else None
            ),
        )

        p = _PendingCommand(cb)
        p.future.add_done_callback(lambda _: self._process_command_finished(p, ignore_errors))
        self._writer.pending_commands.put(p)
        LOG.debug("write_commands(%s)", p)
        return p.future

    def _process_command_finished(self, pending_command, ignore_errors):
        try:
            self._writer.inflight_commands.remove(pending_command)
        except ValueError:
            LOG.warning("Tried to remove %s even though it was already removed.", pending_command)

        if pending_command.future.cancelled():
            LOG.debug("Command cancelled: %s", pending_command)
        else:
            ex = pending_command.future.exception()
            if ex is not None:
                # TODO: more with this; maybe let the user respond to this
                LOG.exception(
                    "A command submission failed!", exc_info=pending_command.future.exception()
                )
            else:
                LOG.debug("Command finished: %s", pending_command)

    async def main_writer(self):
        """
        Main coroutine for submitting commands.
        """
        LOG.info("Writer loop for party %s is starting...", self.party)
        ledger_fut = ensure_future(self._pool.ledger())

        client = self._client_fut.result()  # type: LedgerClient
        metadata = ledger_fut.result()  # type: LedgerMetadata
        validator = ValidateSerializer(self.parent.lookup)

        self._writer.pending_commands.start()

        # Asynchronously iterate over all pending commands that a user has (or will) send.
        # This asynchronous loop "blocks" when pending_commands is empty and is "woken up"
        # immediately when new data is added to it. The loop terminates when the pending_commands
        # ServiceQueue is stopped.
        async for p in self._writer.pending_commands:
            LOG.debug("Sending a command: %s", p)
            command_payloads = []  # type: List[Tuple[_PendingCommand, Sequence[CommandPayload]]]

            if p.future.done():
                # PendingCommand instances that are already marked as done have either been marked
                # as a failure or cancelled by the caller. Do NOT send the corresponding Ledger API
                # command because the PendingCommand() has effectively been aborted.
                continue

            self._writer.inflight_commands.append(p)
            try:
                defaults = CommandDefaults(
                    default_party=self.party,
                    default_ledger_id=metadata.ledger_id,
                    default_workflow_id=None,
                    default_application_id=self._config.application_name,
                    default_command_id=None,
                )
                cps = p.command.build(defaults)
                if cps:
                    commands = await metadata.package_loader.do_with_retry(
                        lambda: [
                            replace(cp, commands=validator.serialize_commands(cp.commands))
                            for cp in cps
                        ]
                    )
                    command_payloads.append((p, commands))
                    await submit_command_async(client, p, commands)
                else:
                    # This is a "null command"; don't even bother sending to the server. Immediately
                    # resolve the future successfully and discard
                    if not p.future.done():
                        p.future.set_result(None)
            except Exception as ex:
                LOG.exception("Tried to send a command and failed!")
                p.notify_read_fail(ex)

        LOG.info("Writer loop for party %s is winding down.", self.party)

        # After the pending command list is fully empty (and never to be filled again), wait for
        # all outstanding commands.
        done, pending = await wait(
            [pc.future for pc in self._writer.inflight_commands],
            timeout=5,
            return_when=ALL_COMPLETED,
        )

        if pending:
            LOG.warning(
                "Writer loop for party %s has NOT fully finished, "
                "but will be terminated anyway (%d futures still pending).",
                self.party,
                len(pending),
            )
        else:
            LOG.info("Writer loop for party %s is finished.", self.party)

    def writer_idle(self):
        return not bool(self._writer.pending_commands) and not self._writer.inflight_commands

    def stop_writer(self):
        """
        Prohibit future command submissions from being accepted.
        """
        self._writer.pending_commands.stop()

    # endregion


@dataclass
class _PartyClientReaderState:
    offset: Optional[str] = field(default=None)


@dataclass
class _PartyClientWriterState:
    pending_commands: "ServiceQueue[_PendingCommand]" = field(default_factory=ServiceQueue)
    inflight_commands: "List[_PendingCommand]" = field(default_factory=list)


class _PendingCommand:
    """
    Track the status of a set of commands in-flight.
    """

    __slots__ = ("command", "command_ids", "future")

    def __init__(self, command: CommandBuilder):
        self.command = safe_cast(CommandBuilder, command)
        self.command_ids = None  # type: Optional[Set[str]]
        self.future = get_event_loop().create_future()

    def notify_write(self, command_ids: Collection[str]):
        self.command_ids = set(command_ids)

    def notify_read_done(self, command_id: str):
        """
        Trigger an appropriate response given the receipt of a transaction.
        """
        if self.command_ids is not None:
            if command_id not in self.command_ids:
                return

            self.command_ids.discard(command_id)
            if not self.command_ids:
                # the command is finished
                if not self.future.done():
                    self.future.set_result(None)
                return

    def notify_read_fail(self, ex: Exception):
        if self.future.done():
            # The user may have cancelled the command or otherwise terminated the Future; trying
            # to set an exception on a done Future will itself throw an exception, so the best
            # we can do is log
            LOG.exception(
                "An exception was received and reported to a Future that has already been aborted! "
                "This cannot be reported in a normal way.",
                ex,
            )
        else:
            self.future.set_exception(ex)

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return id(self)

    def __repr__(self):
        with StringIO() as buf:
            buf.write("<_PendingCommand(command=")
            buf.write(format(self.command, "c"))
            if self.command_ids is not None:
                buf.write(", command_ids=")
                buf.write(format(self.command_ids))
            buf.write(", future=")
            buf.write(format(self.future))
            buf.write(">")
            return buf.getvalue()


def submit_command_async(
    client: LedgerClient, p: "_PendingCommand", commands: Sequence[CommandPayload]
) -> Awaitable[None]:
    """
    Submit a command asynchronously.

    The returned future is resolved when all commands have successfully completed their network
    operations, but not necessarily when the ledger has committed the results of that command to the
    ledger.

    :param client:
    :param p:
    :param commands:
    :return:
    """
    coros = []
    command_ids = []

    for payload in commands:
        coro = None
        command_ids.append(payload.command_id)
        try:
            coro = ensure_future(client.commands(payload))

            LOG.info(
                "cmd submit: party %s, workflow_id %r, command_id %r, %s: %r",
                payload.party,
                payload.workflow_id[0:7],
                payload.command_id[0:7],
                n_things(len(payload.commands), "commands"),
                reprlib.repr(payload.commands),
            )
        except Exception as ex:
            # TODO: consider what to do in this case
            p.notify_read_fail(ex)

        if coro is not None:
            coros.append(coro)

    p.notify_write(command_ids)
    if len(coros) == 0:
        return completed(None)
    elif len(coros) == 1:
        return coros[0]
    else:
        return named_gather("SubmitCommandAsync()", *coros, return_exceptions=True)


def _find_party_config(party_configs: "Collection[dict]", party: Party) -> "Optional[dict]":
    """
    Look within a config dictionary (as specified by FlatConfig) for a configuration object that is
    specific to this party.
    """
    if party_configs is not None:
        for party_config in party_configs:
            local_party = party_config.get("party")
            if local_party == party:
                return party_config
    return None



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Callable, Optional
import warnings

from ..model.types import Type as OldType
from ..util.typing import safe_cast
from ._base import T
from .daml_lf_1 import PrimType, Type


def var(var_: str) -> "Type":
    warnings.warn("dazl.damlast.types.var is deprecated; use dazl.damlast.daml_types.var instead.")
    from .daml_types import var as _var

    return _var(var_)


def match_prim_type(
    prim_type: "Type.Prim",
    on_unit: "Callable[[], T]",
    on_bool: "Callable[[], T]",
    on_int: "Callable[[], T]",
    on_decimal: "Callable[[], T]",
    on_text: "Callable[[], T]",
    on_datetime: "Callable[[], T]",
    on_timedelta: "Callable[[], T]",
    on_party: "Callable[[], T]",
    on_list: "Callable[[Type], T]",
    on_update: "Callable[[Type], T]",
    on_scenario: "Callable[[Type], T]",
    on_date: "Callable[[], T]",
    on_contract_id: "Callable[[Type], T]",
    on_optional: "Callable[[Type], T]",
    on_arrow: "Callable[[Type, Type], T]",
    on_text_map: "Callable[[Type], T]",
    on_numeric: "Callable[[int], T]",
    on_any: "Callable[[], T]",
    on_type_rep: "Callable[[], T]",
    on_gen_map: "Callable[[Type, Type], T]",
) -> "T":
    if prim_type.prim == PrimType.UNIT:
        return on_unit()
    elif prim_type.prim == PrimType.BOOL:
        return on_bool()
    elif prim_type.prim == PrimType.INT64:
        return on_int()
    elif prim_type.prim == PrimType.DECIMAL:
        return on_decimal()
    elif prim_type.prim == PrimType.TEXT or prim_type.prim == PrimType.CHAR:
        return on_text()
    elif prim_type.prim == PrimType.TIMESTAMP:
        return on_datetime()
    elif prim_type.prim == PrimType.RELTIME:
        return on_timedelta()
    elif prim_type.prim == PrimType.PARTY:
        return on_party()
    elif prim_type.prim == PrimType.LIST:
        return on_list(prim_type.args[0])
    elif prim_type.prim == PrimType.UPDATE:
        return on_update(prim_type.args[0])
    elif prim_type.prim == PrimType.SCENARIO:
        return on_scenario(prim_type.args[0])
    elif prim_type.prim == PrimType.DATE:
        return on_date()
    elif prim_type.prim == PrimType.CONTRACT_ID:
        return on_contract_id(prim_type.args[0])
    elif prim_type.prim == PrimType.OPTIONAL:
        return on_optional(prim_type.args[0])
    elif prim_type.prim == PrimType.ARROW:
        return on_arrow(prim_type.args[0], prim_type.args[1])
    elif prim_type.prim == PrimType.TEXTMAP:
        return on_text_map(prim_type.args[0])
    elif prim_type.prim == PrimType.NUMERIC:
        return on_numeric(prim_type.args[0].nat)
    elif prim_type.prim == PrimType.ANY:
        return on_any()
    elif prim_type.prim == PrimType.TYPE_REP:
        return on_type_rep()
    elif prim_type.prim == PrimType.GENMAP:
        return on_gen_map(prim_type.args[0], prim_type.args[1])
    else:
        raise ValueError(f"undefined PrimType: {prim_type}")


def get_old_type(daml_type: "Type") -> "OldType":
    from ..model.types import UnsupportedType

    return safe_cast(Type, daml_type).Sum_match(
        _old_type_var,
        _old_type_con,
        _old_type_prim,
        _old_type_syn,
        _old_forall_type,
        lambda tuple_: UnsupportedType("Tuple"),
        lambda nat: UnsupportedType("Nat"),
        lambda _: UnsupportedType("Syn"),
    )


def _old_type_var(var_: "Type.Var") -> "OldType":
    from ..model.types import TypeApp, TypeVariable

    core_type = TypeVariable(var_.var)
    return (
        TypeApp(core_type, [get_old_type(var_) for var_ in var_.args]) if var_.args else core_type
    )


def _old_type_con(con: "Type.Con") -> "OldType":
    from ..model.types import TypeApp, TypeReference

    core_type = TypeReference(con.tycon)
    return TypeApp(core_type, [get_old_type(var) for var in con.args]) if con.args else core_type


def _old_type_prim(prim: "Type.Prim") -> "OldType":
    from ..model.types import UnsupportedType

    return match_prim_type(
        prim,
        _old_scalar_type_unit,
        _old_scalar_type_bool,
        _old_scalar_type_integer,
        _old_scalar_type_decimal,
        _old_scalar_type_text,
        _old_scalar_type_datetime,
        _old_scalar_type_reltime,
        _old_scalar_type_party,
        _old_list_type,
        _old_update_type,
        lambda *args: UnsupportedType("Scenario"),
        _old_scalar_type_date,
        _old_scalar_contract_id_type,
        _old_optional_type,
        lambda *args: UnsupportedType("Arrow"),
        _old_textmap_type,
        _old_scalar_type_numeric,
        _old_scalar_type_any,
        _old_scalar_type_type_rep,
        _old_genmap_type,
    )


def _old_type_syn(tysyn: "Type.Syn") -> "OldType":
    from ..model.types import TypeApp, TypeReference

    core_type = tysyn.tysyn
    return (
        TypeApp(TypeReference(core_type), [get_old_type(arg) for arg in tysyn.args])
        if tysyn.args
        else core_type
    )


def _old_forall_type(forall: "Type.Forall") -> "OldType":
    from ..model.types import ForAllType

    return ForAllType(forall.vars, get_old_type(forall.body))


def _old_scalar_type_unit() -> "OldType":
    from ..model.types import SCALAR_TYPE_UNIT

    return SCALAR_TYPE_UNIT


def _old_scalar_type_bool() -> "OldType":
    from ..model.types import SCALAR_TYPE_BOOL

    return SCALAR_TYPE_BOOL


def _old_scalar_type_integer() -> "OldType":
    from ..model.types import SCALAR_TYPE_INTEGER

    return SCALAR_TYPE_INTEGER


def _old_scalar_type_decimal() -> "OldType":
    from ..model.types import SCALAR_TYPE_DECIMAL

    return SCALAR_TYPE_DECIMAL


def _old_scalar_type_numeric(nat: int = 10) -> "OldType":
    from ..model.types import SCALAR_TYPE_NUMERIC

    return SCALAR_TYPE_NUMERIC


def _old_scalar_type_text() -> "OldType":
    from ..model.types import SCALAR_TYPE_TEXT

    return SCALAR_TYPE_TEXT


def _old_scalar_type_datetime() -> "OldType":
    from ..model.types import SCALAR_TYPE_DATETIME

    return SCALAR_TYPE_DATETIME


def _old_scalar_type_reltime() -> "OldType":
    from ..model.types import SCALAR_TYPE_RELTIME

    return SCALAR_TYPE_RELTIME


def _old_scalar_type_party() -> "OldType":
    from ..model.types import SCALAR_TYPE_PARTY

    return SCALAR_TYPE_PARTY


def _old_scalar_type_any() -> "OldType":
    from ..model.types import SCALAR_TYPE_ANY

    return SCALAR_TYPE_ANY


def _old_list_type(arg: "Type") -> "OldType":
    from ..model.types import ListType

    return ListType(get_old_type(arg))


def _old_update_type(arg: "Type") -> "OldType":
    from ..model.types import UpdateType

    return UpdateType(get_old_type(arg))


def _old_scalar_type_date() -> "OldType":
    from ..model.types import SCALAR_TYPE_DATE

    return SCALAR_TYPE_DATE


def _old_scalar_contract_id_type(arg: "Type") -> "OldType":
    from ..model.types import ContractIdType

    return ContractIdType(get_old_type(arg))


def _old_optional_type(arg: "Type") -> "OldType":
    from ..model.types import OptionalType

    return OptionalType(get_old_type(arg))


def _old_textmap_type(value_type: "Type") -> "OldType":
    from ..model.types import TextMapType

    return TextMapType(get_old_type(value_type))


def _old_scalar_type_type_rep(arg: "Optional[Type]" = None) -> "OldType":
    from ..model.types import TypeRefType, UnsupportedType

    if arg is not None:
        return TypeRefType(get_old_type(arg))
    else:
        return UnsupportedType("TypeRef")


def _old_genmap_type(key_type: "Type", value_type: "Type") -> "OldType":
    from ..model.types import GenMapType

    return GenMapType(get_old_type(key_type), get_old_type(value_type))
